#include <PSX/File/Processor/cd_file_processor.hpp>
#include <PSX/AutoLBA/auto_lba_declaration.hpp>
#include <PSX/GPU/gpu.hpp>
#include <PSX/GPU/gpu_primitives.hpp>
#include <stdio.h>

enum LBA {
    __jabyengine_start_lba_request
    __jabyengine_request_lba_for(FONT, "ASSETS/FONT.BIN"),
    __jabyengine_request_lba_for(ICON, "ASSETS/ICON.BIN"),
    __jabyengine_end_lba_request
};

// Some default values for the objects
static constexpr auto TriangleColor = JabyEngine::GPU::Color24(0x0, 0xFF, 0xFF);
static constexpr auto TriangleArea  = JabyEngine::GPU::AreaI16({0, 0}, {64, 64});
static constexpr auto TriangleTPage = JabyEngine::GPU::TPage(320, 0, JabyEngine::GPU::SemiTransparency::B_Half_add_F_Half, JabyEngine::GPU::TexturePageColor::$4bit);
static constexpr auto TriangleClut  = JabyEngine::GPU::PageClut(320, 511);

static constexpr auto RectangleColor = JabyEngine::GPU::Color24(0x80, 0x80, 0xFF);
static constexpr auto RectangleArea  = JabyEngine::GPU::AreaI16({0, TriangleArea.size.height}, {80, 80});
static constexpr auto RectangleTPage = JabyEngine::GPU::TPage(320, 256, JabyEngine::GPU::SemiTransparency::B_Half_add_F_Half, JabyEngine::GPU::TexturePageColor::$4bit);
static constexpr auto RectangleClut  = JabyEngine::GPU::PageClut(320, 510);

static constexpr auto LineColor = JabyEngine::GPU::Color24(0xFF, 0x0, 0x0);

static constexpr const auto triangle1 = JabyEngine::GPU::POLY_F3({
    {TriangleArea.position.x, TriangleArea.position.y},
    {TriangleArea.size.width, TriangleArea.size.height},
    {TriangleArea.position.x, TriangleArea.size.height}}, 
    TriangleColor
);
static constexpr const auto triangle2 = JabyEngine::GPU::POLY_FT3({
    {TriangleArea.position.x, TriangleArea.position.y}, 
    {TriangleArea.size.width, TriangleArea.position.y}, 
    {TriangleArea.size.width, TriangleArea.size.height}},{
    // Texture
    {TriangleArea.position.x, TriangleArea.position.y}, 
    {TriangleArea.size.width, TriangleArea.position.y}, 
    {TriangleArea.size.width, TriangleArea.size.height}},
    TriangleTPage,        
    TriangleClut,
    JabyEngine::GPU::Color24::Grey()
);
static constexpr const auto triangle3 = JabyEngine::GPU::POLY_G3({
    {triangle1.vertex0.move(TriangleArea.size.width, 0), JabyEngine::GPU::Color24::Red()},
    {triangle1.vertex1.move(TriangleArea.size.width, 0), JabyEngine::GPU::Color24::Green()},
    {triangle1.vertex2.move(TriangleArea.size.width, 0), JabyEngine::GPU::Color24::Blue()}}
);
static constexpr const auto triangle4 = JabyEngine::GPU::POLY_GT3({
    {triangle2.vertex0.move(TriangleArea.size.width, 0), triangle2.page0, JabyEngine::GPU::Color24::Red()},
    {triangle2.vertex1.move(TriangleArea.size.width, 0), triangle2.page1, JabyEngine::GPU::Color24::Blue()},
    {triangle2.vertex2.move(TriangleArea.size.width, 0), triangle2.page2, JabyEngine::GPU::Color24::Green()}},
    TriangleTPage,
    TriangleClut
);

static constexpr const auto rectangle1 = JabyEngine::GPU::POLY_F4(RectangleArea, RectangleColor);
static constexpr const auto rectangle2 = JabyEngine::GPU::POLY_FT4({
    RectangleArea.position.move(RectangleArea.size.width, 0), RectangleArea.size}, {0, 0}, 
    RectangleTPage, 
    RectangleClut, 
    JabyEngine::GPU::Color24::Grey()
);
static constexpr const auto rectangle3 = JabyEngine::GPU::POLY_G4(
    {RectangleArea.position.move(RectangleArea.size.width*2, 0), RectangleArea.size}, {
    JabyEngine::GPU::Color24::Red(),
    JabyEngine::GPU::Color24::Blue(),
    JabyEngine::GPU::Color24::Green(),
    JabyEngine::GPU::Color24::White()});
static constexpr const auto rectangle4 = JabyEngine::GPU::POLY_GT4(
    {RectangleArea.position.move(RectangleArea.size.width*3, 0), RectangleArea.size}, {0, 0},
    RectangleTPage, 
    RectangleClut, {
    JabyEngine::GPU::Color24::Red(),
    JabyEngine::GPU::Color24::Blue(),
    JabyEngine::GPU::Color24::Green(),
    JabyEngine::GPU::Color24::White()}
);
static constexpr const auto rectangle5 = JabyEngine::GPU::POLY_GT4(
    {RectangleArea.position.move(0, RectangleArea.size.height), RectangleArea.size}, {0, 0},
    RectangleTPage, 
    RectangleClut, {
    JabyEngine::GPU::Color24::Red(),
    JabyEngine::GPU::Color24::Blue(),
    JabyEngine::GPU::Color24::Green(),
    JabyEngine::GPU::Color24::White()}
).set_semi_transparent(true);

static constexpr const JabyEngine::GPU::LINE_F line1[] = {
    JabyEngine::GPU::LINE_F::new_line(LineColor, false),
    JabyEngine::GPU::LINE_F::new_point({0,   0}),
    JabyEngine::GPU::LINE_F::new_point({JabyEngine::GPU::Display::Width, JabyEngine::GPU::Display::Height})
};

static void load_assets() {
    static const JabyEngine::CDFile Assets[] = {
        JabyEngine::CDFileBuilder::simple_tim(LBA::FONT, JabyEngine::SimpleTIM(320,   0, 320, 511)),
        JabyEngine::CDFileBuilder::simple_tim(LBA::ICON, JabyEngine::SimpleTIM(320, 256, 320, 510)),
    };

    const auto                  buffer_cfg = JabyEngine::CDFileProcessor::BufferConfiguration::new_default();
    JabyEngine::CDFileProcessor file_processor;

    file_processor.setup(lba, Assets, buffer_cfg);
    while(true) {
        switch(file_processor.process()) {
            case JabyEngine::Progress::InProgress:
                break;

            case JabyEngine::Progress::Done:
                if(!file_processor.next(lba, buffer_cfg)) {
                    return;
                }
                break;
            
            case JabyEngine::Progress::Error:
                printf("Error detected! Aborting load\n");
                return;
        }
    }
    printf("Done loading assets!\n");
}

void main() {
    load_assets();
    printf("Dino: 0x%02X\n", rectangle4.code.value);

    while(true) {
        JabyEngine::GPU::render(triangle1);
        JabyEngine::GPU::render(triangle2);
        JabyEngine::GPU::render(triangle3);
        JabyEngine::GPU::render(triangle4);

        JabyEngine::GPU::render(rectangle1);
        JabyEngine::GPU::render(rectangle2);
        JabyEngine::GPU::render(rectangle3);
        JabyEngine::GPU::render(rectangle4);
        JabyEngine::GPU::render(rectangle5);

        JabyEngine::GPU::render(line1);

        JabyEngine::GPU::swap_buffers_vsync(2);
    }
}
__declare_lba_header(LBA);